import logging
import os
import subprocess
import warnings

from sqlalchemy import exc as sa_exc

import pytest
from s_code_project_updater.commands import ScodeProjectUpdater
from schema import create_session, Project
from support.logging import get_console_logger

from .test_projects import \
    ScodeTestProject, ScienceTestProject, AlmaTestProject, get_author_pst_ids

_LOG = get_console_logger("scode_project_updater_tests", logging.DEBUG)
_UPDATE_COMMAND = 'update_sproj'
PROFILE = 'local'

class TestUpdater:
    ''' Exercises ScodeProjectUpdater '''
    @pytest.fixture(autouse=True, scope='function')
    def install_test_data(self):
        os.environ['CAPO_PROFILE'] = PROFILE
        self.initialize_test_data()
        yield
        self.remove_test_data()

    def test_dry_run_does_not_update(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        try:
            new_title = 'this is the new title'
            assert fake_project.title != new_title
            args = [
                '-C', project_code,
                '-P', PROFILE,
                '-T', new_title,
                '--dry'
                ]
            updated = ScodeProjectUpdater(args=args).update_project()
            # nothing should have been updated
            assert fake_project.title == updated.title
            assert fake_project.abstract == updated.abstract
            assert len(fake_project.authors) == len(updated.authors)
        except SystemExit as exc:
            pytest.fail(f'unexpected failure with return code {exc.code}')
            raise
        except Exception as exc:
            pytest.fail(f'{project_code}: {exc}')
            raise

    def test_project_code_only_fetches(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        args = [
            '-C', project_code,
            '-P', PROFILE,
        ]

        updated = None
        try:
            updated = ScodeProjectUpdater(args=args).update_project()
        except SystemExit as exc:
            pytest.fail(f'unexpected failure with return code {exc.code}')
            raise

        assert updated != None

        assert fake_project.title == updated.title
        assert fake_project.abstract == updated.abstract
        assert len(fake_project.authors) == len(updated.authors)
        count = 0
        for orig_author in fake_project.authors:
            for author in updated.authors:
                if author.username == orig_author.username:
                    count += 1
                    break
        assert len(fake_project.authors) == count

    def test_updates_abstract_only(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        new_abstract = "Well, here's another nice mess you've gotten us into, Ollie"
        assert fake_project.abstract != new_abstract
        args = [
            '-C', project_code,
            '-P', PROFILE,
            '-A', new_abstract,
        ]
        try:
            updated = ScodeProjectUpdater(args=args).update_project()
            # only abstract should have been updated;
            # all else should be same
            assert fake_project.title == updated.title
            assert new_abstract == updated.abstract
            assert len(fake_project.authors) == len(updated.authors)
        except SystemExit as exc:
            pytest.fail(f'unexpected failure; return code = {exc.code}')
            raise

    def test_updates_abstract_and_title(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        new_abstract = "I think you ought to know I'm feeling very depressed"
        new_title = 'A Survey of the Mattresses of Sqornshellous Zeta'
        assert fake_project.abstract != new_abstract
        assert fake_project.title != new_title
        args = [
            '-C', project_code,
            '-P', PROFILE,
            '-A', new_abstract,
            '-T', new_title,
        ]
        try:
            updated = ScodeProjectUpdater(args=args).update_project()
            assert new_title == updated.title
            assert new_abstract == updated.abstract
            assert len(fake_project.authors) == len(updated.authors)
        except SystemExit as exc:
            pytest.fail(f'unexpected failure; exit code = {exc.code}')
            raise

    def test_adds_new_abstract_deletes_author(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        new_project = Project(project_code=project_code,
                              title=fake_project.title,
                              abstract=fake_project.abstract)
        new_abstract = "First there is a mountain, then there is no " \
                        "mountain, then there is"
        assert new_abstract != fake_project.abstract
        new_project.abstract = new_abstract
        original_authors = fake_project.authors.copy()
        assert 4 == len(original_authors)
        last_author = original_authors[3]
        new_authors = original_authors[:3]
        assert len(original_authors) - 1 == len(new_authors)
        new_project.authors = new_authors
        args = [
            '-C', project_code,
            '-P', PROFILE,
            '-A', new_abstract,
            '-I',
        ]
        for author_id in get_author_pst_ids(new_project):
            args.append(str(author_id))

        updated = None
        try:
            updated = ScodeProjectUpdater(args=args).update_project()
            assert updated is not None
        except SystemExit as exc:
            pytest.fail(f'unexpected failure; return code = {exc.code}')
            raise

        assert fake_project.abstract != updated.abstract
        assert fake_project.title == updated.title
        assert len(original_authors) - 1 == len(updated.authors)
        authors_updated = last_author in updated.authors
        assert not authors_updated
        count = 0
        for orig_author in original_authors[:3]:
            for new_author in updated.authors:
                if new_author.username == orig_author.username:
                    count += 1
                    break
        assert len(new_authors) == count

    def test_output_is_as_expected(self):
        fake_project = ScodeTestProject().project
        project_code = fake_project.project_code
        args = [
            '-C', project_code,
            '-P', PROFILE,
        ]
        updater = ScodeProjectUpdater(args=args)
        updater.update_project()
        output = updater.get_project_info()
        assert output is not None
        assert ('Title: ' + fake_project.title) in output
        assert ('Abstract: ' + fake_project.abstract) in output
        pst_ids = [str(id) for id in get_author_pst_ids(fake_project)]
        pst_id_str = ' '.join(pst_ids)
        assert 'Authors: ' + pst_id_str in output

    def test_copes_with_single_pi(self):
        project = ScodeTestProject().project
        args = ['-P', PROFILE, '-C', project.project_code, '-I', '4686']
        try:
            updated = ScodeProjectUpdater(args=args).update_project()
            assert 1 == len(updated.authors)
        except SystemExit as ex:
            pytest.fail(f'update failed with exit code {ex.code}')
            raise

    def test_alma_project_is_rejected(self):
        project_code = '2018.A.00062.S'
        args = ['-P', PROFILE, '-C', project_code,
                '-T', 'Physics at High Angular Resolution in Nearby Galaxies: '
                      'The Local Galaxy Inventory Continued']

        with pytest.raises(SystemExit) as exc:
            ScodeProjectUpdater(args=args).update_project()
            assert 2 == exc.code

    def test_update_failure_returns_expected_code(self):
        result = FailingUpdater().update_project()
        assert isinstance(result, SystemExit)
        assert 5 == result.code

    """ The following test should be moved to another test case, 
        where we'll use a bash script, via subprocess.call(), to create an 
        appropriate env and execute pytest.
    """
    @pytest.mark.skip('pytest passes only in IJ; '
                      'fails when run from command line'
                      'due to import errors')
    def test_command_line_returns_expected_codes(self):
        ''' We simulate execution from the command line
            and confirm that errors result in the appropriate
            return codes.

        '''

        # minimum required arguments -- profile & project -- omitted
        return_code = CommandLineUpdaterLauncher([]).run()
        assert return_code == 2

        project_code = ScodeTestProject().project.project_code

        # profile not specified
        args = ['-C', project_code,]
        return_code = CommandLineUpdaterLauncher(args).run()
        assert return_code == 2

        # project code not specified
        args = ['-P', PROFILE]
        assert CommandLineUpdaterLauncher(args).run() == 2

        # profile value missing
        args = ['-P', '-C', project_code]
        return_code = CommandLineUpdaterLauncher(args).run()
        assert return_code == 2

        # project code missing
        args = ['-P', PROFILE, '-C']
        assert CommandLineUpdaterLauncher(args).run() == 2

        # bad project code
        args = ['-P', PROFILE, '-C', 'bogus']
        assert CommandLineUpdaterLauncher(args).run() == 3

        # bad profile
        args = ['-P', 'not_a_profile', '-C', project_code]
        assert CommandLineUpdaterLauncher(args).run() == 1

        # missing title as last argument
        args = ['-P', PROFILE, '-C', project_code, '-T']
        assert CommandLineUpdaterLauncher(args).run() == 2

        # missing title as first argument
        args = ['-T', '-P', PROFILE, '-C', project_code,]
        assert CommandLineUpdaterLauncher(args).run() == 2

        # nonexistent investigator
        args = ['-P', PROFILE, '-C', project_code, '-I', '-22']
        assert CommandLineUpdaterLauncher(args).run() == 4


    ### UTILITIES ###

    def initialize_test_data(self):
        ''' Insert test data into archive database for use in tests '''
        session = create_session('SDM')
        num_commits = num_found = 0
        try:
            with warnings.catch_warnings():
                # Suppress SQLAlchemy warnings
                warnings.simplefilter("ignore", category=sa_exc.SAWarning)

                fake_projects = [ScodeTestProject().project,
                                 ScienceTestProject().project,
                                 AlmaTestProject().project]
                try:
                    for fake_project in fake_projects:
                        project_code = fake_project.project_code
                        existing = session.query(Project) \
                            .filter(Project.project_code ==
                                    project_code) \
                            .first()
                        if existing is not None:
                            num_found += 1
                            session.delete(existing)
                            session.commit()
                        session.add(fake_project)
                        session.commit()
                        num_commits += 1

                    if num_commits < num_found:
                        pytest.fail(f'{num_found} fake projects were found '
                                    f'and deleted, but {num_commits} were '
                                    f'added and committed')
                except Exception as exc:
                    pytest.fail(f'{exc}')
                    raise
        finally:
            session.close()

    def remove_test_data(self):
        ''' Get rid of the test data we inserted. '''
        session = create_session('SDM')
        try:
            with warnings.catch_warnings():
                # Suppress SQLAlchemy warnings
                warnings.simplefilter("ignore", category=sa_exc.SAWarning)

                fake_projects = [ScodeTestProject().project,
                                 ScienceTestProject().project,
                                 AlmaTestProject().project]
                for fake_project in fake_projects:
                    project_code = fake_project.project_code
                    existing = session.query(Project) \
                        .filter(Project.project_code ==
                                project_code) \
                        .first()
                    if existing is not None:
                        session.delete(existing)
                        session.commit()

                # confirm removal
                found = session.query(Project) \
                    .filter(Project.project_code.like('%_TEST_PROJECT')) \
                    .first()
                if found is not None:
                    pytest.fail('test projects were not removed')
        except Exception as exc:
            pytest.fail(f'{exc}')
        finally:
            session.close()

class FailingUpdaterHelper:
    ''' for use in testing update failure '''

    @pytest.fixture()
    def update_project(self):
        return SystemExit(5)

class FailingUpdater:
    ''' for use in testing update failure '''
    def __init__(self):
        self.helper = FailingUpdaterHelper()

    def update_project(self):
        return SystemExit(5)

class CommandLineUpdaterLauncher:
    ''' Simulates execution of script from command line.
        This works when tests are run from within iJ
        but not when pytest is execuated at the command line.
     '''

    def __init__(self, args: list):
        self.args = [_UPDATE_COMMAND]
        for arg in args:
            self.args.append(str(arg))
        _LOG.info(f'{self.args}')

    def run(self):
        ''' launch updater in a subprocess
            @:returns directory listing
        '''

        args = self.args
        try:
            proc = subprocess.run(args,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT,
                                  timeout=60,
                                  check=False,
                                  bufsize=1,
                                  universal_newlines=True)
            self.stdout = proc.stdout
            if proc.returncode:
                print(f'{self.stdout}')
            return proc.returncode
        except Exception as exc:
            _LOG.error(f'{exc}')
            return exc.returncode
