import glob
import json
import logging
import os
import re

from pycapo import *
from pyramid.view import view_config

from . import ResponseMsg, def_name

logger = logging.getLogger(__name__)


@view_config(route_name='get_casa_version_list', renderer='jsonp')
def get_casa_version_list(request):
    r""" Build a list of available CASA versions and return them to the caller.

        The front end allows the user to select a version of CASA to use when processing their archive requests, but
        only from a managed list of versions, found in a directory we define in CAPO. This directory is mirrored to
        both sites nightly, and consists of symlinks to installs (relerases, pre-release and tests) of CASA+Pipeline,
        plus an additional link that denotes which version of CASA is recommended for the archive: if that link is
        missing we presume the most recent version is the default.

        If the CAPO property edu.nrao.archive.archiveIface.allowDevCasaVersions we will show pre-releases and tests,
        if they are there. Otherwise we only show releases.

        The lists will be constructed as a sorted list of version numbers in descending order, with the 'default'
        set as the first element in the list.  If the soft links are present, we resolve the path and version,
        then remove that numbered revision from the list below it, so we aren't presenting duplicates.
    :param request: brings in the build profile
    :return:
        casa_version_list: a list of CASA versions, sorted with descending version number and the 'default' at the
                            top of the list.
        response_msg: A ResponseMsg object containing a success/failure flag, an integer success/failure code, and
                        a message describing the success/failure of the requested operation.
    """

    def casa_version_from_path(path):
        """ Strip off 'casa-release-|casa-prerelease-|casa-test-' and '-elN' or '.elN', where N is a number
        :param path: A path to a CASA version
        :return: The version number derived from the path name
        """
        name = os.path.split(path)[1]
        cleaned_name = re.sub(r'[\.\-]el[0-9]$', r'', re.sub(r'casa-(.*release-|.*test-)', '', name))
        return cleaned_name

    version_list = []
    vla_procedures_file = None
    vla_intents_file = None
    alma_procedures_file = None
    alma_intents_file = None
    default_link='default'
    try:
        # Fetch assorted settings from CAPO
        capo_profile = request.session['capo_profile']
        c = CapoConfig(profile=capo_profile)
        casa_root = c.get('edu.nrao.archive.archiveIface.casa_root')
        vla_procedures_file = c.get('edu.nrao.archive.workflow.config.CiplWorkflowSettings.casaProceduresFileName')
        vla_intents_file = c.get('edu.nrao.archive.workflow.config.CiplWorkflowSettings.casaIntentsFileName')
        alma_procedures_file = c.get(
            'edu.nrao.archive.workflow.config.CiplWorkflowSettings.almaStandardProceduresFileName')
        alma_intents_file = c.get('edu.nrao.archive.workflow.config.CiplWorkflowSettings.almaStandardIntentsFileName')
        allow_dev_casa = True if c.get('edu.nrao.archive.archiveIface.allowDevCasaVersions') == 'true' else False

        try:
            os.listdir(casa_root)
        except:
            raise Exception('Failed to locate CASA root')

        if allow_dev_casa:
            prefix_match = r'^casa-(pipeline-)?(prerelease|release|test)-.*'
        else:
            prefix_match = r'^casa-(pipeline-)?release-.*'

        # Start by building a list of version/path dictionaries, extracting the version name from the directory
        # names in our CASA root. The assumption here is that all these CASA installs are also pipeline installs.
        sorted_dir = list(reversed(sorted(os.listdir(casa_root))))
        usable_versions = [dict(version=casa_version_from_path(o),
                                path=os.path.join(casa_root, o))
                           for o in sorted_dir
                           if (os.path.isdir(os.path.join(casa_root, o)) or
                               os.path.islink(os.path.join(casa_root, o))) and
                           re.match(prefix_match, o)]

        # Now let's look for the 'default' soft link in the directory.  If we find it, we resolve the link to get the
        # version name, make it the first element in our return list and append our version list from above, with the
        # 'default' (resolved) version removed, so we don't show it as a duplicate.
        if default_link in sorted_dir:
            default_version_path = os.path.realpath(os.path.join(casa_root, default_link))
            version_list = [dict(version=casa_version_from_path(os.path.split(default_version_path)[1]),
                                 path=default_version_path)]
            dedup_default = [ver for ver in usable_versions
                             if not ver['version'] == version_list[0]['version']]
            version_list.extend(dedup_default)
        else:
            # if the default wasn't found, we copy the version list we've already build and tag the highest revision
            # as the 'default'.  We copy the version list from above rather than simple assignment because we're
            # going to modify the list.
            version_list = [dict(version=ver['version'], path=ver['path']) for ver in usable_versions]
        version_list[0]['version'] += ' (recommended)'

        response_msg = ResponseMsg(True, ResponseMsg.SUCCESS,
                                   'Found the CASA directory, returning results.')
    except Exception as ex:
        ex_msg = 'Exception caught in {0}.{1}: {2}'.format(__name__, def_name(), ex)
        logger.info(ex_msg)
        response_msg = ResponseMsg(False, ResponseMsg.HARD_ERROR, ex_msg)

    # logger.info(response_msg.as_dict())
    return json.dumps(dict(casa_version_list=version_list,
                           vla_procedures_file=vla_procedures_file,
                           vla_intents_file=vla_intents_file,
                           alma_procedures_file=alma_procedures_file,
                           alma_intents_file=alma_intents_file,
                           response_msg=response_msg.as_dict()))


